import time
import socket
import struct
import datetime
from py_logging.cestel_logging import log, log_exception
import logging
import os

from generic_module import Module

class CanModule(Module):
  def __init__(self, name, udp_rx_ip, udp_rx_port, udp_tx_ip, udp_tx_port, udp_e_ip, udp_e_port, interval, root_path, relevant_for_listeners, ap_fwver, board_ids):
    Module.__init__(self, name, root_path)
    self.udp_rx_ip = udp_rx_ip
    self.udp_rx_port = udp_rx_port
    
    self.udp_tx_ip = udp_tx_ip
    self.udp_tx_port = udp_tx_port
    
    self.udp_e_ip = udp_e_ip
    self.udp_e_port = udp_e_port

    self.interval = interval
    self.relevant_for_listeners = relevant_for_listeners #=u, t, ut, 0
    
    self.vdesc = ["unknown","analog_12V","analog_-12V","analog_5V","analog_-5V","digital_12V"]
    self.tdesc = ["unknown", "T1__C", "T2__C", "T3__C", "T4__C", "Ti__C"]
    self.bname = ["unknown", "CAZ", "CAT"]
    
    self.ap_fwver = ap_fwver
    self.board_ids = board_ids
    self.rx_s = None
    
  def generate_path(self, mod):
    pth = self.sites_path
    pth = pth + "/" + self.sname + "/ext"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + mod
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/can"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.name 
    if not os.path.exists(pth):
      os.mkdir(pth)
    return pth

  def can_msg(self, id, data):
    if len(data) > 8:
        raise ValueError("invalid data")
    d = bytearray(36) # initializes elements to zero
    d[1] = 0x24 # = 36
    d[3] = 0x80
    d[21] = len(data)
    d[24:28] = struct.pack('>i', id)  # > indicates big endian
    d[28:28+len(data)] = data
    return d

  def brd_cmd(self, board_id, cmd, par1, par2):
    d = bytearray(8)
    d[0] = board_id
    d[1] = cmd
    d[2:4] = struct.pack('<i', par1)[:2]  # struct.pack returns 4 bytes, but we only care about first 2. < indicates little endian
    d[4:8] = struct.pack('<i', par2)
    txsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    txsock.settimeout(1)
    txsock.sendto(self.can_msg(0xc0,d), (self.udp_tx_ip, self.udp_tx_port))
    
  def dict2confstr(self, section, d):
    r = '[' + section + ']' + '\n'
    for k,v in d.items():
      r += k + '=' + v + '\n'
    return r
    
  def send_data(self, data):
    conf_str_data = self.dict2confstr("mydevice", data)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
      sock.sendto(conf_str_data.encode(), (self.udp_e_ip, self.udp_e_port))
    finally:
      sock.close()
    
  def bind_udp(self):
    self.rx_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)    
    self.rx_s.settimeout(1)
    try:
      self.rx_s.bind((self.udp_rx_ip, self.udp_rx_port))  
      return 0
    except:
      log_exception(self.logger, 'Failed to bind socket')
      return 1
  
  def run(self):
    if self.bind_udp() == 0:
      start = time.time()
      temp_dict = dict()
      volt_dict = dict()
      while True:
        if self.end:
          try:
            self.rx_s.close()
          except:
            log_exception(self.logger, 'Failed to close:')
          self.alive = False
          log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
          return
        try:
          #data1 being binary string
          #data being byte array
          data, addr = self.rx_s.recvfrom(64)
          data = bytearray(data)  # Does nothing in Python3 as return value is already bytearray.
          dlc = data[21]
          cid = struct.unpack('!I', data[24:28])[0]
          cdt = data[28:36]
          #VOLTAGE, do nothing
          if cid == 0x90:
            vid = data[28]
            volt_dict[self.vdesc[int(vid)]] = format(struct.unpack('f', data[29:33])[0],'.2f')
            #volt_dict["ts"] = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
          #TEMPERATURE
          elif cid == 0x80:
            tid = data[28]
            temp_dict[self.tdesc[int(tid)]] = format(struct.unpack('f', data[29:33])[0],'.2f')
            temp_dict["ts"] = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
          #VERSION
          elif (cid >= 0xc1)and(cid <= 0xcf):
            board_id = cdt[0]  # Since we have a bytearray, values are represented as numbers, so cast to int is irrelevant (and not Python3 compatible).
            par_id = cdt[1]
            if par_id == 1: # fwver
              fwver_major = cdt[3]
              fwver_minor = cdt[2]
            ver = '{0}.{1}'.format(fwver_major, fwver_minor)
            for key, mod in self.downstream_modules_dict.items():
              if board_id == 1:
                mod.set_upstream_info(self.get_name(), "caz_firmware_version", ver)
                #mod.set_caz_fw(ver)
              elif board_id == 2:
                mod.set_upstream_info(self.get_name(), "cat_firmware_version", ver)
                #mod.set_cat_fw(ver)
          else:
            log(self.logger, 'unk ID ' + format(cid, '03x') + ' DATA ' + ' '.join(format(x, '02x') for x in cdt[:dlc]), sev=logging.ERROR)
        except socket.timeout:
          pass
        except:
          log_exception(self.logger, 'Failed to obtain info:')
        for key, mod in self.downstream_modules_dict.items():
          for volt_key, val in volt_dict.items():
            if volt_key != "unknown" and volt_key != "ts":
              #print("voltage!")
              mod.set_upstream_info(self.get_name(), volt_key, str(val))
        if time.time() - start > self.interval and "t" in self.relevant_for_listeners.lower():
          #maybe create method
          if self.sname != "":
            pth = self.generate_path("temperature")
            for key, mod in self.downstream_modules_dict.items():
              mod.set_upstream_info(self.get_name(), "temperature_path", pth)
            date = datetime.datetime.now().strftime('%Y-%m-%d')
            fd = open(pth + "/" + date + ".csv", "a+")
            header = "\t".join(self.tdesc[1:])
            header = "ts" + "\t" + header
            if not os.path.exists(pth + "/" + date + ".csv") or os.path.getsize(pth + "/" + date + ".csv") == 0:
              fd.write(header + "\n")
            tm = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
            to_write = tm + "\t"
            for el in self.tdesc[1:]:
              try:
                to_write = to_write + "\t" + temp_dict[el]
              except:
                to_write = to_write + "\t" + "-237"
            fd.write(to_write + "\n")
            fd.close()
          #-------------------
          self.send_data(temp_dict)
          for board_id in self.board_ids:
            self.brd_cmd(board_id, self.ap_fwver, 0, 0)
          start = time.time()
    try:
      self.rx_s.close()
    except:
      log_exception(self.logger, 'Failed to close:')
    self.alive = False
    return
 
